#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["get_batches"]

from itertools import (
    chain,
    islice,
)
from typing import (
    Iterable,
    List,
    TypeVar,
)

T = TypeVar("T")


def get_batches(iterable: Iterable[T], *, batch_size: int) -> Iterable[List[T]]:
    assert batch_size > 0

    source_iter = iter(iterable)
    while True:
        # return consequent slices of `batch_size` elements
        slices = islice(source_iter, batch_size)
        try:
            first_from_slice = next(slices)
        except StopIteration:
            # but if there's nothing to return in last slice, close generator instead of returning empty list
            return
        yield list(chain([first_from_slice], slices))
