__all__ = ["ANONYMOUS_API_TOKEN_CONTENT", "DO_NOT_TRACK_GIT_REPOSITORY"]

ANONYMOUS_API_TOKEN_CONTENT = (
    "eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYXBwLm5lcHR1bmUuYWkiLCJhcGlfdXJsIjoiaHR0cHM6Ly9hcHAubmVwdHVuZS"
    "5haSIsImFwaV9rZXkiOiJiNzA2YmM4Zi03NmY5LTRjMmUtOTM5ZC00YmEwMzZmOTMyZTQifQo="
)
DO_NOT_TRACK_GIT_REPOSITORY = "DO_NOT_TRACK_GIT_REPOSITORY"
